@@ -25,17 +25,18 @@ const MAGIC_4CHN:[u8; 4] =['4' as u8, 'C' as u8, 'H' as u8, 'N' as u8];
 const MAGIC_6CHN:[u8; 4] =['6' as u8, 'C' as u8, 'H' as u8, 'N' as u8];
 const MAGIC_8CHN:[u8; 4] =['8' as u8, 'C' as u8, 'H' as u8, 'N' as u8];
 
-///
-/// Periods from http://greg-kennedy.com/tracker/modformat.html
-///
-///          C    C#   D    D#   E    F    F#   G    G#   A    A#   B
-/// Octave 1: 856, 808, 762, 720, 678, 640, 604, 570, 538, 508, 480, 453
-/// Octave 2: 428, 404, 381, 360, 339, 320, 302, 285, 269, 254, 240, 226
-/// Octave 3: 214, 202, 190, 180, 170, 160, 151, 143, 135, 127, 120, 113
-///
-/// Octave 0:1712,1616,1525,1440,1357,1281,1209,1141,1077,1017, 961, 907
-/// Octave 4: 107, 101,  95,  90,  85,  80,  76,  71,  67,  64,  60,  57
-///
+//
+// Periods from http://greg-kennedy.com/tracker/modformat.html
+//
+//          C    C#   D    D#   E    F    F#   G    G#   A    A#   B
+// Octave 1: 856, 808, 762, 720, 678, 640, 604, 570, 538, 508, 480, 453
+// Octave 2: 428, 404, 381, 360, 339, 320, 302, 285, 269, 254, 240, 226
+// Octave 3: 214, 202, 190, 180, 170, 160, 151, 143, 135, 127, 120, 113
+//
+// Octave 0:1712,1616,1525,1440,1357,1281,1209,1141,1077,1017, 961, 907
+// Octave 4: 107, 101,  95,  90,  85,  80,  76,  71,  67,  64,  60,  57
+//
+/// Lookup table of amiga periods
 pub static PERIODS: &'static [u16] = &[
     1712,1616,1525,1440,1357,1281,1209,1141,1077,1017, 961, 907,
     856,  808, 762, 720, 678, 640, 604, 570, 538, 508, 480, 453,
@@ -44,11 +45,14 @@ pub static PERIODS: &'static [u16] = &[
 	107,  101,  95,  90,  85,  80,  76,  71,  67,  64,  60,  57,
 ];
 
+/// Lookup table of note names
 pub static NOTE_NAMES: &'static [&'static str] = &["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"];
 
+/// Lookup table for 4 bit delta packing
 static DELTA_4BIT: &'static [u8] = &[0x00, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 
 									 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff];
 
+/// Lookup table for effect names
 pub static EFFECT_NAMES: &'static [&'static str] = &[
 		// Normal commads 0-F
 		"Arpeggio", "Slide up", "Slide down", "Slide to note",
@@ -474,8 +478,23 @@ impl PTModule {
 			} // for rows
 		} // for patterns
 	}
+
+	pub fn remove_breaks_on_last_rows(&mut self) {
+		for pattern in &mut self.patterns {
+			let i = pattern.rows.len()-1;
+			for channel in &mut pattern.rows[i].channels {
+				let effect = (channel.effect & 0xf00) >> 8;
+				if effect == 0xd {
+					// remove it
+					channel.effect = 0;
+				}
+			} // for channels
+		} // for patterns
+	}
+
 }
 
+/// read data until the array is full
 fn read_all(reader: &mut dyn Read, data: &mut [u8]) -> io::Result<usize> {
 	let mut pos = 0;
 	while pos < data.len() {
@@ -490,12 +509,14 @@ fn read_all(reader: &mut dyn Read, data: &mut [u8]) -> io::Result<usize> {
 	Ok(pos)
 }
 
+/// write all data in the array
 fn write_all(writer: &mut dyn Write, data: &[u8]) -> io::Result<()> {
 	let n = {writer.write_all(&data)}?;
 	
 	Ok(n)
 }
 
+/// read a 0 padded string stored as a sequence of u8
 fn read_0_padded_string(reader: &mut dyn Read, len: usize) -> Result<String, io::Error> {
 	let mut data = vec![0u8; len];
 	read_all(reader,&mut data)?;
@@ -507,6 +528,7 @@ fn read_0_padded_string(reader: &mut dyn Read, len: usize) -> Result<String, io:
 	Ok(str)
 }
 
+/// write a string as zero padded sequence of u8
 fn write_0_padded_string(writer: &mut dyn Write, str: &String, len: usize) -> io::Result<()> {
 	let mut data = vec![0u8; len];
 	let str_buf = str.as_bytes();
@@ -520,6 +542,7 @@ fn write_0_padded_string(writer: &mut dyn Write, str: &String, len: usize) -> io
 	Ok(n)
 }
 
+/// read a big endian u16
 fn read_big_endian_u16(reader: &mut dyn Read) -> Result<u16, io::Error> {
 	let mut data_arr = [0u8; 2];
 	read_all(reader,&mut data_arr)?;
@@ -533,6 +556,7 @@ fn read_big_endian_u16(reader: &mut dyn Read) -> Result<u16, io::Error> {
 	Ok(data)
 }
 
+/// write a big endian u16
 fn write_big_endian_u16(writer: &mut dyn Write, val: u16)  -> io::Result<()> {
 	let mut data = [0u8; 2];
 	data[0] = (val >> 8) as u8;
@@ -543,6 +567,7 @@ fn write_big_endian_u16(writer: &mut dyn Write, val: u16)  -> io::Result<()> {
 	Ok(n)
 }
 
+/// read a single u8
 fn read_u8(reader: &mut dyn Read) -> Result<u8, io::Error> {
 	let mut data = [0u8; 1];
 	
@@ -551,7 +576,7 @@ fn read_u8(reader: &mut dyn Read) -> Result<u8, io::Error> {
 	Ok(data[0])
 }
 
-
+/// write a single u8
 fn write_u8(writer: &mut dyn Write, val: u8)  -> io::Result<()> {
 	let mut data = [0u8; 1];
 	data[0] = val;
@@ -663,33 +688,33 @@ pub fn read_mod(reader: &mut dyn Read, ignore_file_size_check: bool) -> Result<P
 	let mut module = PTModule::new();
 
 	// First read 20 bytes songname
-	module.name = read_0_padded_string(reader, 20)?;
+	module.name = read_0_padded_string(reader, 20).expect("Failed to read module name");
 
 	// Read all sample info
 	// TODO Handle 15 sample files
 	for _i in 0..DEFAULT_NUMBER_OF_SAMPLES {
 		let mut si = SampleInfo::new();
 		// Sample name
-		si.name = read_0_padded_string(reader, 22)?;
-		si.length = read_big_endian_u16(reader)?;
+		si.name = read_0_padded_string(reader, 22).expect("Failed to read sample name");
+		si.length = read_big_endian_u16(reader).expect("Failed to read sample length");
 		// Finetune
-		si.finetune = read_u8(reader)?;
+		si.finetune = read_u8(reader).expect("Failed to read finetune");
 		// Volume
-		si.volume = read_u8(reader)?;
+		si.volume = read_u8(reader).expect("Failed to read volume");
 		// Repeat start
-		si.repeat_start = read_big_endian_u16(reader)?;
+		si.repeat_start = read_big_endian_u16(reader).expect("Failed to read repeat start");
 		// Repeat length
-		si.repeat_length = read_big_endian_u16(reader)?;
+		si.repeat_length = read_big_endian_u16(reader).expect("Failed to read repeat length");
 
 		module.sample_info.push(si);
 	}
 	
 	// Songlength
-	module.length = read_u8(reader)?;
+	module.length = read_u8(reader).expect("Failed to read module length");
 	// nt_restart
-	module.nt_restart = read_u8(reader)?;
+	module.nt_restart = read_u8(reader).expect("Failed to read nt restart");
 	// Song positions
-	read_all(reader,&mut module.positions.data)?;
+	read_all(reader,&mut module.positions.data).expect("Failed to read positions");
 	// M.K.
 	read_all(reader,&mut module.mk)?;
 	if module.mk != MAGIC_MK &&
@@ -721,7 +746,7 @@ pub fn read_mod(reader: &mut dyn Read, ignore_file_size_check: bool) -> Result<P
 		for row in &mut pattern.rows {
 			for channel in &mut row.channels {
 				let mut data  = [0u8;4];
-				read_all(reader,&mut data)?;
+				read_all(reader,&mut data).expect("Failed to read pattern data");
 				channel.sample_number = (data[0] & 0xf0) | ((data[2] & 0xf0) >> 4);
 				channel.period = (((data[0] & 0x0f) as u16) << 8) | (data[1] as u16);
 				channel.effect = (((data[2] & 0x0f) as u16) << 8) | (data[3] as u16);
@@ -735,7 +760,12 @@ pub fn read_mod(reader: &mut dyn Read, ignore_file_size_check: bool) -> Result<P
 		if si.length > 0 {
 			let length_in_bytes = si.length * 2;
 			let mut data = vec![0u8; length_in_bytes as usize];
-			read_all(reader,&mut data)?;
+			match read_all(reader,&mut data) {
+				Err(e) => {
+					println!("Warning: Not enough sample data, file is likely corrupt. Error: '{}'",e);
+				},
+				_ => ()
+			};
 			si.data = data;
 		}
 	}
@@ -1174,6 +1204,267 @@ pub fn read_p61(reader: &mut dyn Read) -> Result<PTModule, PTMFError> {
 	Ok(module)
 }
 
+/// Read an Amiga ProTracker file packed with The Player 6.1 and write some info about it
+pub fn analyze_p61(reader: &mut dyn Read) -> Result<PTModule, PTMFError> {
+	let mut module = PTModule::new();
+	module.num_channels = DEFAULT_NUMBER_OF_CHANNELS_PER_ROW;
+	
+	// Read the entire file to a Vec<u8>
+	// since the format uses a lot of offsets
+	// back and forth in the file.
+	let mut data:Vec<u8> = Vec::new();
+	let data_size = reader.read_to_end(&mut data)?;
+	if data.len() != data_size {
+		return Err(PTMFError::Parse(format!("Failed to read all bytes {} {}", data.len(), data_size)));	
+	}
+	
+	let mut pos = 0;
+
+	if data[0] == 'P' as u8 && 
+		data[1] == '6' as u8 &&
+		data[2] == '1' as u8 &&
+		data[3] == 'A' as u8 {
+		println!("has header");
+		// Remove header
+		data.remove(0);
+		data.remove(0);
+		data.remove(0);
+		data.remove(0);
+	}
+
+	let sample_offset = ((data[pos] as u16) << 8) | data[pos+1] as u16;
+	let num_patterns = data[pos+2];
+	println!("number of patterns {}",num_patterns);
+	let num_samples = data[pos+3];
+	pos += 4;
+	
+	let is_delta_8_bit = num_samples & 0x80 == 0x80;
+	let is_delta_4_bit = num_samples & 0x40 == 0x40;
+	let num_samples = num_samples & 0b00111111;
+
+	println!("number of samples {} 8 bit delta packing {} 4 bit delta packing {}",num_samples,is_delta_8_bit,is_delta_4_bit);
+	
+	let mut unpacked_samples_length:u32 = 0; 
+	if is_delta_4_bit {
+		for _ in 0..4 {
+			unpacked_samples_length = (unpacked_samples_length << 8) + data[pos] as u32;
+			pos += 1;
+		}
+		println!("unpacked sample length {}",unpacked_samples_length);
+	}
+	
+	println!("samples:");
+	let mut sample_start = sample_offset as usize;
+	for i in 0..num_samples as usize {
+		module.sample_info.push(SampleInfo::new());
+		let sample_length = ((data[pos] as u16) << 8) | data[pos+1] as u16;
+		let finetune = data[pos+2];
+		let volume = data[pos+3];
+		let mut repeat_start = ((data[pos+4] as u16) << 8) | data[pos+5] as u16;
+		let mut repeat_length = 0;
+		if repeat_start == 0xffff {
+			repeat_start = 0;
+		} else {
+			repeat_length = sample_length - repeat_start;
+		}
+		
+		println!("  sample length {} finetune {} volume {} repeat start {} repeat length {}",sample_length,finetune,volume,repeat_start,repeat_length);
+
+		let signed_sample_length = sample_length as i16;
+		if signed_sample_length < 0 && signed_sample_length >= -31 {
+			// This means that the sample uses the same sample data as
+			// another sample
+			let sample_index = (-1 * signed_sample_length - 1) as usize;
+			
+			let length = module.sample_info[sample_index].length;
+			let data = module.sample_info[sample_index].data.clone();
+
+			// Fix repeat length
+			if repeat_start == 0xffff {
+				repeat_start = 0;
+			} else {
+				repeat_length = length - repeat_start;
+			}
+
+			println!("   using the same sample data as {}",sample_index);
+			println!("   sample length {} finetune {} volume {} repeat start {} repeat length {}",length,finetune & 0xf,volume,repeat_start,repeat_length);
+			
+			let si = &mut module.sample_info[i];
+			si.length = length;
+			si.finetune = finetune & 0x0f;
+			si.volume = volume;
+			si.repeat_start = repeat_start;
+			si.repeat_length = repeat_length;
+			si.data = data;
+
+		} else {
+		
+			let is_sample_delta_4_bit = finetune & 0x80 == 0x80;
+		
+			let si = &mut module.sample_info[i];
+			si.length = sample_length;
+			si.finetune = finetune & 0x0f;
+			si.volume = volume;
+			si.repeat_start = repeat_start;
+			si.repeat_length = repeat_length;
+
+			// 4-bit delta can be enabled/disabled per sample, bit 7 of finetune (0x80)
+			// It is possible to combine 8-bit and 4-bit delta
+			if is_delta_4_bit && is_sample_delta_4_bit {
+				println!("   sample data is 4 bit delta packed");
+				let sample_end = sample_start + sample_length as usize; // Packed length is half of unpacked length
+				let mut delta:u8 = 0;
+				for i in sample_start..sample_end {
+					let hi = ((data[i] & 0xf0) >> 4) as usize;
+					let lo = (data[i] & 0x0f) as usize;
+					let subhi = DELTA_4BIT[hi];
+					delta = (Wrapping(delta) - Wrapping(subhi)).0;
+					si.data.push(delta);
+					let sublo = DELTA_4BIT[lo];
+					delta = (Wrapping(delta) - Wrapping(sublo)).0;
+					si.data.push(delta);				
+				}
+				
+				// Move to next sample
+				sample_start = sample_end;
+				
+			} else if is_delta_8_bit {
+				println!("   sample data is 8 bit delta packed");
+				let sample_end = sample_start + (sample_length as usize) * 2;
+				let mut delta:u8 = data[sample_start];
+				// First byte get copied unmodified
+				si.data.push(delta);
+				
+				for i in sample_start+1..sample_end {
+					delta = (Wrapping(delta) - Wrapping(data[i])).0;
+					si.data.push(delta);
+				}
+				
+				// Move to next sample
+				sample_start = sample_end;
+			
+			} else {
+				let sample_end = sample_start + (sample_length as usize) * 2;
+				let sample_data = &data[sample_start..sample_end];
+				si.data = sample_data.to_vec();
+
+				// Move to next sample
+				sample_start = sample_end;
+				
+			}
+		}
+		
+		// Move to next sample
+		pos = pos+6;
+	}
+	
+	let mut offset_pos = pos;
+	let mut pattern_offsets:Vec<usize> = Vec::new();
+	for _ in 0..num_patterns {
+		module.patterns.push(Pattern::new(DEFAULT_NUMBER_OF_ROWS_PER_PATTERN, DEFAULT_NUMBER_OF_CHANNELS_PER_ROW));
+		for _ in 0..DEFAULT_NUMBER_OF_CHANNELS_PER_ROW {
+			let offset = (((data[pos] as u16) << 8) | data[pos+1] as u16) as usize;
+			pattern_offsets.push(offset);
+			
+			pos += 2;
+		}
+	}
+	
+	println!("positions:");
+	print!("  ");
+	let mut length = 0;
+	loop {
+		let position = data[pos];
+		if position == 0xff {
+			break;
+		}
+		module.positions.data[length] = position;
+		print!("{} ", position);
+		
+		length += 1;
+		pos += 1;
+	}
+	println!();
+	
+	pos += 1;
+
+	module.length = length as u8;
+	
+	let pattern_start_offset = pos;
+
+	println!("offsets:");
+	for i in 0..num_patterns {
+		print!("  {}: ",i);
+		for _ in 0..DEFAULT_NUMBER_OF_CHANNELS_PER_ROW {
+			let offset = (((data[offset_pos] as u16) << 8) | data[offset_pos+1] as u16) as usize + pattern_start_offset;
+
+			print!("{:02x}({})  ",offset,offset);
+			
+			offset_pos += 2;
+		}
+		println!();
+	}
+
+
+	
+	// Pop removes from the end of Vec
+	pattern_offsets.reverse();
+	
+	for pattern_number in 0..num_patterns as usize {
+	
+		let mut row_number = [0usize; DEFAULT_NUMBER_OF_CHANNELS_PER_ROW];
+		let mut current_pos = [0usize; DEFAULT_NUMBER_OF_CHANNELS_PER_ROW];
+		
+		for channel_number in 0..DEFAULT_NUMBER_OF_CHANNELS_PER_ROW as usize {
+			row_number[channel_number] = 0;
+			current_pos[channel_number] = pattern_start_offset + pattern_offsets.pop().unwrap();
+		}
+		
+		let mut truncate_pos = DEFAULT_NUMBER_OF_ROWS_PER_PATTERN;
+		loop {
+			
+			// We always need to decode the channel with the lowest row number first
+			// to see if it contains a pattern break or jump
+			
+			let mut current_channel = 0;
+			let mut lowest = DEFAULT_NUMBER_OF_ROWS_PER_PATTERN;
+			for channel_number in 0..DEFAULT_NUMBER_OF_CHANNELS_PER_ROW as usize {
+				if row_number[channel_number] < lowest {
+					lowest = row_number[channel_number];
+					current_channel = channel_number;
+				}
+			}
+			
+			if lowest >= truncate_pos {
+				break;
+			}
+
+//			println!("P {} C {} R {} P {:X}",pattern_number,current_channel,row_number[current_channel],current_pos[current_channel]);
+			let eop = decode_p61_row(&data, &mut current_pos[current_channel], pattern_number, &mut row_number[current_channel], current_channel, &mut module);				
+			if eop {
+				// This will make sure we exit the while loop early
+				// but still process any remaining channels
+				truncate_pos = row_number[current_channel];
+			}
+						
+		} // loop
+		
+		// If we exit the while loop early above, make sure that we clear out any data
+		// that we parsed but shouldn't really be there
+		for row_number in truncate_pos..DEFAULT_NUMBER_OF_ROWS_PER_PATTERN as usize {
+			for channel_number in 0..DEFAULT_NUMBER_OF_CHANNELS_PER_ROW as usize {
+				let channel = &mut module.patterns[pattern_number].rows[row_number].channels[channel_number];
+				channel.period = 0;
+				channel.sample_number = 0;
+				channel.effect = 0;
+			}
+		}		
+	}
+
+	Ok(module)
+}
+
+/// encode one channel onw row in The Player format
 fn encode_p61_channel(channel: &Channel) -> Vec<u8> {
 	let mut data:Vec<u8> = Vec::new();
 
@@ -1194,7 +1485,7 @@ fn encode_p61_channel(channel: &Channel) -> Vec<u8> {
 	if cmd == 0xE {
 		// E commands are special, handled below
 		match param1 {
-			0 => effect = (effect & 0xff) | ((param2 & 1) * 2),
+			0 => effect = (effect & 0xf00) | ((param2 & 1) * 2),
 			1 | 2 | 9 | 0xA | 0xB | 0xD | 0xE if param2 == 0 => effect = 0, // No effect if empty paramters,
 			3 | 4 | 5 | 6 | 7 | 8 | 0xF => (),
 			0xC if param2 == 0 => effect = 0xC00, // Replace with empty volume
@@ -1368,18 +1659,27 @@ pub fn write_p61(writer: &mut dyn Write, module: &PTModule) -> Result<(),PTMFErr
 	let mut workmodule = module.clone();
 
 	// Prepare all data first
-	workmodule.truncate_samples();
-	// To generate binary identical files,
-	// we have to leave duplicate samples
-	//workmodule.remove_duplicate_samples();
-	workmodule.remove_unused_samples();
-	workmodule.truncate_patterns();
-	workmodule.remove_duplicate_breaks();
+
+	// First remove patterns
 	// to generate binary identical files
 	// we have to leave duplicate patterns
 	//workmodule.remove_duplicate_patterns();
 	workmodule.remove_unused_patterns();
 
+	// Then remove samples
+	workmodule.truncate_samples();
+	// To generate binary identical files,
+	// we have to leave duplicate samples
+	//workmodule.remove_duplicate_samples();
+	workmodule.remove_unused_samples();
+
+	// Can only do this after removing samples
+	// since some samples might be used in
+	// the parts that gets removed
+	// D commands on last row of patterns should be removed
+	workmodule.remove_breaks_on_last_rows();
+	workmodule.truncate_patterns();
+	workmodule.remove_duplicate_breaks();
 
 	// The Player uses 2 passes
 	// First pass: 
@@ -1477,8 +1777,11 @@ pub fn write_p61(writer: &mut dyn Write, module: &PTModule) -> Result<(),PTMFErr
 			}
 			packed_pattern_offsets[pattern_idx][channel_idx] = packed_stream_cursor.position() as usize;
 
+//			println!("=>>> channel_idx {} pattern_idx {} stream_cursor.position() {} packed_stream_cursor.position() {}",channel_idx,pattern_idx,stream_cursor.position(),packed_stream_cursor.position());
+
 			let channel_end_offset = pattern_offsets[pattern_idx+1][channel_idx] as u64;
 			while stream_cursor.position() < channel_end_offset {
+//				println!("stream_cursor.position() {} packed_stream_cursor.position() {}",stream_cursor.position(),packed_stream_cursor.position());
 				if first {
 					first = false;
 					// First time is special and can not be compressed
@@ -1488,89 +1791,184 @@ pub fn write_p61(writer: &mut dyn Write, module: &PTModule) -> Result<(),PTMFErr
 
 				// save current positions
 				let current_packed_stream_pos = packed_stream_cursor.position();
+				let current_stream_pos = stream_cursor.position();
 
-				// Write current value to stream
-				p61_putdata(&mut stream_cursor, &mut packed_stream_cursor)?;
+				let mut tmp_area = Vec::new();
+				let mut tmp_cursor = Cursor::new(&mut tmp_area);
 
-				// get it back for comparison
-				packed_stream_cursor.set_position(current_packed_stream_pos);
-				let current_value = p61_fetchdata(&mut packed_stream_cursor)?;
+				// fetch next value from uncompressed stream
+				tmp_cursor.set_position(0);
+				p61_putdata(&mut stream_cursor, &mut tmp_cursor)?;
+				tmp_cursor.set_position(0);
+				let first_src_value = p61_fetchdata(&mut tmp_cursor)?;
+
+//				println!(" first_src_value {:x}",first_src_value);
 
 				let mut search_cursor = Cursor::new(packed_stream_cursor.get_ref());
 				search_cursor.set_position(0);
-				let mut src_cursor = Cursor::new(stream_cursor.get_ref());
 
-				let mut longest_match_pos = 0;
-				let mut longest_match_count = 0;
-				let mut consumed_pos = 0;
-				let mut found_match = false;
+				let mut best_match = false;
+				let mut best_match_offset = 0;
+				let mut best_match_length = 0;
+				let mut best_match_length_bytes = 0;
+				let mut best_match_cursor = 0;
 				while search_cursor.position() < current_packed_stream_pos {
-					let mut tmp_area = Vec::new();
-					let mut tmp_cursor = Cursor::new(&mut tmp_area);
-	
-					let search_position = search_cursor.position();
 
-					let existing_value = p61_fetchdata(&mut search_cursor)?;
+					// save start of search position
+					// this is used to calculate the number of
+					// compressed bytes
+					let start_search_pos = search_cursor.position();
 
-					if current_value == existing_value {
-						// Match
-						let mut match_count = 1;
+					// fetch one value from compressed stream
+					let first_compressed_value = p61_fetchdata(&mut search_cursor)?;
 
-						// Keep searching
-						src_cursor.set_position(stream_cursor.position());
+					if first_src_value != first_compressed_value {
+						// this does not match, keep searching
+						continue;
+					}
 
-						while src_cursor.position() <  channel_end_offset && 
-								search_cursor.position() < current_packed_stream_pos {
-							// decode next src
+					// save this position and reset stream here
+					// before next search
+					let next_search_pos = search_cursor.position();
+
+					// ok, the first one matched
+					let mut src_cursor = Cursor::new(stream_cursor.get_ref());
+					src_cursor.set_position(stream_cursor.position());
+
+					if src_cursor.position() >= channel_end_offset {
+						// end of unpacked data for this channel and pattern
+						// no point in continuing searching
+						break;
+					}
+
+					if search_cursor.position() >= current_packed_stream_pos {
+						// end of packed data
+						// no point in continuing searching
+						break;
+					}
+
+					// fetch next value from uncompressed stream
+					tmp_cursor.set_position(0);
+					p61_putdata(&mut src_cursor, &mut tmp_cursor)?;
+					tmp_cursor.set_position(0);
+					let second_src_value = p61_fetchdata(&mut tmp_cursor)?;
+
+					// fetch one value from compressed stream
+					let second_compressed_value = p61_fetchdata(&mut search_cursor)?;
+
+					if second_src_value != second_compressed_value {
+						// No match
+						// Reset search position and try again
+						search_cursor.set_position(next_search_pos);
+						continue;
+					}
+
+					// when we get here, two values have matched
+					let mut match_length = 1;
+
+					if src_cursor.position() < channel_end_offset &&
+					search_cursor.position() < current_packed_stream_pos {
+
+						while src_cursor.position() < channel_end_offset &&
+							search_cursor.position() < current_packed_stream_pos {
+
+							// fetch next value from uncompressed stream
 							tmp_cursor.set_position(0);
+							let src_position_if_no_match = src_cursor.position();
 							p61_putdata(&mut src_cursor, &mut tmp_cursor)?;
 							tmp_cursor.set_position(0);
-							let src_value = p61_fetchdata(&mut tmp_cursor)?;
+							let next_src_value = p61_fetchdata(&mut tmp_cursor)?;
 
-							let search_value = p61_fetchdata(&mut search_cursor)?;
+							// fetch one value from compressed stream
+//							let search_position_if_no_match = search_cursor.position();
+							let next_compressed_value = p61_fetchdata(&mut search_cursor)?;
 
-							if src_value == search_value {
-								match_count += 1;
-							} else {
-								src_cursor.set_position(src_cursor.position()-4);
+							if next_src_value != next_compressed_value {
+								// no match, abort
+								// but first undo cursor
+
+								// The Player lets the pointer for the compressed stream
+								// move too far
+								//search_cursor.set_position(search_position_if_no_match);
+
+								// src_cursor is not used after this
+								// but lets reset it anyway
+								src_cursor.set_position(src_position_if_no_match);
 								break;
 							}
+
+							// If we get here we have a match
+							// increase counter and keep searching
+							match_length += 1;
 						}
+					}
 
-						if match_count > 1 && match_count > longest_match_count {
-							found_match = true;
-							longest_match_count = match_count;
-							longest_match_pos = search_position;
-							consumed_pos = src_cursor.position();
-						}
+					// now we need to figure out the length in bytes
+					// of the current match
+					let mut match_length_bytes = (search_cursor.position() - start_search_pos) as i64;
+					// pointer is at least 3 bytes
+					match_length_bytes -= 3;
 
+					// The Player seems to do it this way, 
+					// but it is off by 3 bytes?
+					let pointer_size = current_packed_stream_pos - start_search_pos;
+					if pointer_size >= 256 {
+						match_length_bytes -= 1;
+					}
 
-					} // if match
-	
-				} // while search
+//					println!(" potential match at start_search_pos {} match_length_bytes {} match_length {} first_src_value {:x} second_src_value {:x}", start_search_pos, match_length_bytes, match_length, first_src_value, second_src_value);
 
-				if found_match {
+					if match_length_bytes > best_match_length_bytes {
+//						println!(" new best match");
+						best_match = true;
+						best_match_offset = start_search_pos;
+						best_match_length = match_length;
+						best_match_length_bytes = match_length_bytes;
+						best_match_cursor = src_cursor.position();
+
+						// and now keep searching
+					}
+
+					// Reset search position and try again
+					search_cursor.set_position(next_search_pos);
+				} // while search_cursor.position() < current_packed_stream_pos 
+
+				// now we have searched
+				if best_match {
+					// and we have found
 					// Rewind stream
 					packed_stream_cursor.set_position(current_packed_stream_pos);
 					// Jump forward in stream
-					stream_cursor.set_position(consumed_pos);
+					stream_cursor.set_position(best_match_cursor);
 
 					// calc pointer
-					let mut delta =  current_packed_stream_pos - longest_match_pos + 3;
+					let mut delta =  current_packed_stream_pos + 3 - best_match_offset;
 					if delta < 256 {
+//						println!(" write best match 8 bit {:02x} {:02x} {:02x}",0xff,0b01000000 | best_match_length,(delta & 0xff));
 						// 8 bit pointer
 						packed_stream_cursor.write_all(&[0xff])?;
-						packed_stream_cursor.write_all(&[0b01000000 | longest_match_count-1])?;
+						packed_stream_cursor.write_all(&[0b01000000 | best_match_length])?;
 						packed_stream_cursor.write_all(&[(delta & 0xff) as u8])?;
 					} else {
+//						println!(" write best match 16 bit {:02x} {:02x} {:02x} {:02x}",0xff,0b11000000 | best_match_length,((delta >> 8) & 0xff),(delta & 0xff));
 						// 16 bit pointer
 						delta += 1;
 						packed_stream_cursor.write_all(&[0xff])?;
-						packed_stream_cursor.write_all(&[0b11000000 | longest_match_count-1])?;
+						packed_stream_cursor.write_all(&[0b11000000 | best_match_length])?;
 						packed_stream_cursor.write_all(&[((delta >> 8) & 0xff) as u8])?;
 						packed_stream_cursor.write_all(&[(delta & 0xff) as u8])?;
 					}
-				} // if found_match
+				} else {
+//					println!(" no match ");
+					// did not find any mathcing data
+
+					// rewind source stream and write to packed stream
+					stream_cursor.set_position(current_stream_pos);
+					p61_putdata(&mut stream_cursor, &mut packed_stream_cursor)?;
+				}
+
+				// Time to loop again and try to pack next command
+
 			} // stream_cursor.position() < channel_end_offset
 		} // for pattern_idx 
 	} // for channel_idx
@@ -1604,18 +2002,64 @@ pub fn write_p61(writer: &mut dyn Write, module: &PTModule) -> Result<(),PTMFErr
 	// finetune u8, hi bit 0x80 means sample is 4 bit delta packed
 	// volume u8
 	// repeat start u16 or 0xffff u16
-	for si in &workmodule.sample_info {
-		// length in words
-		let sample_length = si.data.len() as u16 / 2;
-		write_big_endian_u16(&mut final_data_cursor, sample_length)?;
-		write_u8(&mut final_data_cursor, si.finetune)?;
-		write_u8(&mut final_data_cursor, si.volume)?;
-		let repeat_start = if si.repeat_length > 1 {
-			si.repeat_start
+	let mut samples_to_write = Vec::new();
+	for i in 0..workmodule.sample_info.len() {
+		let mut same_as_another = false;
+		let mut same_index = 0;
+		{
+			let si = &workmodule.sample_info[i];
+
+			if si.data.len() > 0 {
+				// only compare if sample isn't empty
+				for j in 0..i {
+					// Check if same sample data as another sample
+					let si_other = &workmodule.sample_info[j];
+					if si.data == si_other.data {
+						same_as_another = true;
+						same_index = j;
+						break;
+					}
+				}	
+			}
+		} // end the borrow of si
+
+		let si = &mut workmodule.sample_info[i];
+		if same_as_another {
+			// Sample that points to another samples data
+			let sample_index = !(same_index as u16 + 1) + 1;
+			write_big_endian_u16(&mut final_data_cursor, sample_index)?;
+			write_u8(&mut final_data_cursor, si.finetune)?;
+			write_u8(&mut final_data_cursor, si.volume)?;
+			let repeat_start = if si.repeat_length > 1 {
+				si.repeat_start
+			} else {
+				0xffff
+			};
+			write_big_endian_u16(&mut final_data_cursor, repeat_start)?;
 		} else {
-			0xffff
-		};
-		write_big_endian_u16(&mut final_data_cursor, repeat_start)?;
+			// Normal sample
+			// add this to the list of samples to write
+			samples_to_write.push(i);
+
+			if si.data.len() == 0 {
+				// Empty samples not allowed, must be at least one word
+				si.data.push(0);
+				si.data.push(0);
+			}
+
+			// length in words
+			let sample_length = si.data.len() as u16 / 2;
+			write_big_endian_u16(&mut final_data_cursor, sample_length)?;
+			write_u8(&mut final_data_cursor, si.finetune)?;
+			write_u8(&mut final_data_cursor, si.volume)?;
+			let repeat_start = if si.repeat_length > 1 {
+				si.repeat_start
+			} else {
+				0xffff
+			};
+			write_big_endian_u16(&mut final_data_cursor, repeat_start)?;
+		}
+
 	}
 
 	// for each pattern
@@ -1660,7 +2104,8 @@ pub fn write_p61(writer: &mut dyn Write, module: &PTModule) -> Result<(),PTMFErr
 
 	// for each sample 
 	// sample data
-	for si in &workmodule.sample_info {
+	for idx in samples_to_write {
+		let si = &workmodule.sample_info[idx];
 		// TODO check that sample length is even
 		let mut sample_cursor = Cursor::new(&si.data);
 		copy(&mut sample_cursor, &mut final_data_cursor)?;
@@ -1700,7 +2145,18 @@ mod tests {
 
 		Ok(())
 	}
-	
+
+	#[test]
+	fn test_analyze_p61() -> Result<(),PTMFError> {
+		let basedir = env!("CARGO_MANIFEST_DIR");
+		let p61filename = format!("{}/testdata/{}",basedir, "P61.spiderfunk.mod");
+		let file = File::open(&p61filename)?;
+		let mut reader = BufReader::new(&file);
+		analyze_p61(&mut reader)?;
+
+		Ok(())
+	}
+
     #[test]
     fn test_write_p61_spiderfunk() -> Result<(),PTMFError> {
 		let basedir = env!("CARGO_MANIFEST_DIR");
@@ -1712,7 +2168,6 @@ mod tests {
 		let mut writer = Cursor::new(&mut created);
 
 		write_p61(&mut writer,&mut module)?;
-
 		let p61filename = format!("{}/testdata/{}",basedir, "P61.spiderfunk.mod");
 		let file = File::open(&p61filename)?;
 		let mut reader = BufReader::new(&file);
